from tqdm import tqdm

from Levenshtein import distance as levenshtein_distance
from tqdm import tqdm
from tamil.utf8 import get_letters
from collections import deque
from pprint import pprint, pformat

from .valam import DEFAULT_DICTIONARY_FILES
from .bloom import build_bloom
from .trie import build_trie
from .bktree import build_bktree

import json

import ilakkani.valam as valam

class Thundu:
    def __init__(self, word):
        self.word = word
        self.suggestions = []
        self.correct = False

    def __json__(self):
        return {
            'word' : self.word,
            'suggestions' : self.suggestions,
            'correct_p' : self.correct
        }
    

class Thiruthi:
    def __init__(self,
                 filepaths = DEFAULT_DICTIONARY_FILES,
                 suggestions_count = 5):
        self.bloom   = build_bloom(filepaths, 200000, 0.05)
        self.trie    = build_trie(filepaths)
        self.bktree  = build_bktree(filepaths)

        self.suggestions_count = suggestions_count

    def _thiruthu(self, thundugal):

        for thundu in thundugal:
            if thundu.word in self.bloom:
                thundu.correct = True
                continue

            if self.trie.prefix_exists_p(thundu.word):
                thundu.correct = True
                continue

            thundu.suggestions.extend(
                self.bktree.search(thundu.word, 2)[:self.suggestions_count]
            )

        return thundugal

    def thiruthu(self, words):
        thundugal = [Thundu(i) for i in words]
        return self._thiruthu(thundugal)

    def copy_dicts(self, other):
        self.bloom = other.bloom
        self.trie = other.trie
        self.bktree = other.bktree

class KoappuThiruthi(Thiruthi):
    def __init__(self, filepaths, suggestions_count=5):
        super().__init__(filepaths, suggestions_count)

    def thiruthu(self, ulleedu, veliyeedu):
        for vari in ulleedu:
            #print(vari)
            thundugal = super().thiruthu(vari.split())
            sarivari = ''
            for thundu in thundugal:
                if thundu.correct:
                    sarivari += ' ' + thundu.word
                elif thundu.suggestions:
                    sarivari += ' ' + thundu.word \
                        + '/' \
                        + '/'.join([i[1] for i in thundu.suggestions])
            #print('  ' + sarivari)
            print(sarivari, file=veliyeedu)
            

class JsonThiruthi(Thiruthi):
    def __init__(self, filepaths, suggestions_count=5):
        super().__init__(filepaths, suggestions_count)

    def thiruthu(self, ulleedu_json):
        thundugal = super().thiruthu(ulleedu_json['chorkal'])
        thirutham = [i.__json__() for i in thundugal]
        ulleedu_json['thirutham'] = thirutham
        return ulleedu_json

bloom = None
trie = None
bktree = None

def main_loop():
    global bloom, trie, bktree
    bloom   = build_bloom(DEFAULT_DICTIONARY_FILES, 200000, 0.05)
    trie    = build_trie(DEFAULT_DICTIONARY_FILES)
    bktree  = build_bktree(DEFAULT_DICTIONARY_FILES)

    word = input('> ')
    while word:
        print('இருக்குதா? {}'.format(
            'இருக்கு' if word in bloom  else 'இல்லை'))
        

        print('இருக்குதா? {}'.format(
            'இருக்கு' if trie.prefix_exists_p(get_letters(word)) else 'இல்லை'))
        
        pprint('என்ன என்ன வார்த்தைகளோ?')
        pprint(bktree.search(word, 2))
        
        word = input('> ')

    return

import sys
import argparse

def vaayil():
    parser = argparse.ArgumentParser('thiruthi')
    parser.add_argument('input',
                        default=sys.stdin,
                        type=argparse.FileType('r'),
                        help='input file')

    parser.add_argument('output',
                        default=sys.stdout,
                        type=argparse.FileType('w'),
                        help='output file')

    args = parser.parse_args()
        
    thiruthi = KoappuThiruthi(DEFAULT_DICTIONARY_FILES, suggestions_count=2)
    thiruthi.thiruthu(args.input, args.output)

    
if __name__ == '__main__':

    vaayil()
    """
    parser = argparse.ArgumentParser('thiruthi')
    parser.add_argument('input',
                        default=sys.stdin,
                        type=argparse.FileType('r'),
                        help='input file')

    parser.add_argument('output',
                        default=sys.stdout,
                        type=argparse.FileType('w'),
                        help='output file')

    args = parser.parse_args()
        
    thiruthi = KoappuThiruthi(DEFAULT_DICTIONARY_FILES, suggestions_count=2)
    thiruthi.thiruthu(args.input, args.output)

    json_thiruthi = JsonThiruthi([])
    json_thiruthi.copy_dicts(thiruthi)

    
    #thirutham = json_thiruthi.thiruthu(json.load(open('tharavu/test.json')))
    #pprint(thirutham)
    
    #main_loop()
    """




import os
VER = os.path.realpath(os.path.join(os.path.dirname(__file__), '..'))

XSV_DELIMITER = '\t'
THARAVU_VER = f'{VER}/tharavu'
DEFAULT_DICTIONARY_FILES = [f'{THARAVU_VER}/chorkuviyal.22mar22.tsv.gz', f'{THARAVU_VER}/chorpirappiyal-agarathi.tsv.gz']
CHELLUMAI_ADAVU_FILE = f'{THARAVU_VER}/orungurri-chellumai-cholloadai.tsv'



# -*- coding: utf-8 -*-

from editdistance import eval as levenshtein
from tqdm import tqdm
from tamil.utf8 import get_letters
from collections import deque
from pprint import pprint, pformat
import csv

import ilakkani.utils as utils
from .valam import DEFAULT_DICTIONARY_FILES

class BKTree:
    """
    From https://iq.opengenus.org/burkhard-keller-tree/
    """
    def __init__(self, distance_func):
        self._tree = None
        self._distance_func = distance_func

    def distance_func(self, node, candidate):
        return self._distance_func(node, candidate)
        
    def add(self, node):
        if self._tree is None:
            self._tree = (node, {})
            return

        current, children = self._tree
        while True:
            dist = self.distance_func(node, current)
            target = children.get(dist)
            if target is None:
                children[dist] = (node, {})
                break
            current, children = target

    def search(self, node, radius):
        if self._tree is None:
            return []

        candidates = deque([self._tree])
        result = []
        while candidates:
            candidate, children = candidates.popleft()
            dist = self.distance_func(node, candidate)
            if dist <= radius:
                result.append((dist, candidate))

            low, high = dist - radius, dist + radius
            candidates.extend(c for d, c in children.items()
                              if low <= d <= high)
            
        return sorted(result, key=lambda x: x[0])

class TamilBKTree(BKTree):
    def distance_func(self, node, candidate):
        node, candidate = get_letters(node), get_letters(candidate)
        return super().distance_func(node, candidate)


def build_bktree(filepaths,
                pbarp = False):
    print('building bktree...')
    tree = TamilBKTree(levenshtein)
    for filepath in filepaths:
        print('loading {}...'.format(filepath))
        if pbarp:
            pbar = tqdm.tqdm(utils.openfile(filepath), ncols=100)
        else:
            pbar = utils.openfile(filepath)

        for item in csv.reader(pbar, delimiter='\t'):
            try:
                token, count = item
                if token:
                    tree.add(token)
                    if pbarp:
                        pbar.set_description(token)
            except:
                print('item: ', item)


    return tree

if __name__ == '__main__':

    tree = TamilBKTree(levenshtein)

    for filepath in DEFAULT_DICTIONARY_FILES:
        print('loading {}...'.format(filepath))
        with utils.openfile(filepath) as f:
            csvf = csv.reader(f, delimiter='\t')
            for line in tqdm(csvf):
                token, count = line
                if token:
                    tree.add(token)
                    
    word = input('> ')
    while word:
        pprint('என்ன என்ன வார்த்தைகளோ?')
        pprint(tree.search(word, 2))
        
        word = input('> ')



    with utilsopen('tamil_tree_output.txt', 'w') as of:
        of.write(pformat(tree))



# -*- coding: utf-8 -*-
from pprint import pprint, pformat
import pdb
from tqdm import tqdm
from tamil.utf8 import get_letters
import csv

import ilakkani.utils as utils
from .valam import DEFAULT_DICTIONARY_FILES, XSV_DELIMITER

class Node(object):

    def __repr__(self):
        if self.level < 160:
            return "Node({}, {}, {})\n{} {}".format(
                ''.join(self.value) if self.value else self.value,
                self.count,
                self.is_complete,
                '\t' * self.level,
                self.children,)
        else:
            return ''

    def __str__(self):
        return self.__repr__()

    def __init__(self, value=None, count=1, level=0):
        self.value = value
        self.count = count
        self.children = {}
        self.is_complete = False
        self.level = level
        

class Trie(object):

    def __init__(self):
        self.root = Node(value=None, level=0)

    def __repr__(self):   return self.root.__repr__()
    def __str__(self):    return self.__repr__()

    def add(self, item):
        node, i = self.find_prefix(item)
        if i < len(item):
            #increment count
            j = 0
            tnode = self.root
            while j < i:
                tnode.children[item[j]].count += 1
                tnode = tnode.children[item[j]]
                j += 1
                
            # add new nodes
            while i < len(item):
                #new_node = Node(item[:i+1], count=1, level=i+1)
                new_node = Node(item[i], count=1, level=i+1)
                node.children[item[i]] = new_node
                node = new_node
                i += 1

            node.is_complete = True

    def find_prefix(self, prefix, default=None):
        i = 0
        prev_node = node = self.root
        while i < len(prefix) and node:
            prev_node = node
            node = node.children.get(prefix[i], None)
            i += 1
            
        if i <= len(prefix):
            return prev_node, i-1
        else:
            return self.root, 0
        
    def prefix_exists_p(self, prefix):
        node, index = self.find_prefix(prefix)
        if not node == self.root:
            node = node.children.get(prefix[-1], None)
            if node:
                return node.is_complete

    def get_all_suffixes(self, prefix):

        suffixes = []
        node, level = self.find_prefix(prefix)
        if len(prefix):
            node = node.children[prefix[-1]]
        branches = list([ ('' + k, v) for k,v in node.children.items()])
        while branches:
            prefix, node = branches.pop(0)
            branches = list([ (prefix + k, v) for k,v in node.children.items()]) + branches
            if node.is_complete:
                suffixes.append(prefix)

        return suffixes

    def words(self):
        return self.get_all_suffixes(self.root)

def build_trie(filepaths,
                pbarp = False):
    print('building trie...')
    trie = Trie()
    for filepath in filepaths:
        print('loading {}...'.format(filepath))
        if pbarp:
            pbar = tqdm.tqdm(utils.openfile(filepath), ncols=100)
        else:
            pbar = utils.openfile(filepath)

        for item in csv.reader(pbar, delimiter=XSV_DELIMITER):
            try:
                token, count = item
                if token:
                    trie.add(get_letters(token))
                    if pbarp:
                        pbar.set_description(token)
            except:
                print('item: ', item)


    return trie
    
if __name__ == '__main__':
    trie = Trie()
    trie.add("hell")
    trie.add("hello")
    trie.add("hey")
    trie.add("trie")
    pprint (trie)
    pprint (trie.find_prefix("hey"))
    pprint (trie.get_all_suffixes("h"))
    pprint (trie.prefix_exists_p("hel"))
    pprint (trie.prefix_exists_p("hell"))
    pprint (trie.prefix_exists_p("heyl"))
    pprint (trie.prefix_exists_p("hey"))
    pprint (trie.prefix_exists_p("tri"))
    pprint (trie.prefix_exists_p("trie"))
    pprint (trie.prefix_exists_p("Trie"))

    tamil_trie = Trie()
    for filepath in DEFAULT_DICTIONARY_FILES:
        print('loading {}...'.format(filepath))
        with utils.openfile(filepath) as f:
            csvf = csv.reader(f, delimiter=XSV_DELIMITER)
            for line in tqdm(csvf):
                token, count = line
                if token:
                    tamil_trie.add(get_letters(token))
                    
    word = input('> ')
    while word:
        print('இருக்குதா? {}'.format(
            'இருக்கு' if tamil_trie.prefix_exists_p(get_letters(word)) \
            else 'இல்லை'))
        
        word = input('> ')


    with utils.openfile('tamil_trie_output.txt', 'w') as of:
        of.write(pformat(tamil_trie))



import gzip
import pdb

def openfile(filepath, mode='rt', *args, **kwargs):
    if filepath.endswith('.gz'):
        return gzip.open(filepath, mode, *args, **kwargs)
    else:
        return open(filepath, mode, *args, **kwargs)


def compare_dict(dict1, dict2, debug=0):
    if isinstance(dict1, dict) and isinstance(dict1, dict):
        dict1keys = sorted(dict1.keys())
        dict2keys = sorted(dict2.keys())

        assert dict1keys == dict2keys
        nested_score = []
        for k in dict1keys:
            _score = compare_dict(dict1[k], dict2[k], debug+1)
            nested_score.append(_score)
            
        return all(nested_score)


    else:
        return dict1 == dict2



# -*- coding: utf-8 -*-
from bloom_filter import BloomFilter
from random import shuffle
from tqdm import tqdm
import tqdm

import csv
import ilakkani.utils as utils
from .valam import DEFAULT_DICTIONARY_FILES, XSV_DELIMITER

BLOOMFILTER_SIZE = 200000 #no of items to add
BLOOMFILTER_PROB = 0.05 #false positive probability


def build_bloom(filepaths,
                size  = BLOOMFILTER_SIZE,
                prob  = BLOOMFILTER_PROB,
                pbarp = False):

    print('building bloom filter...')
    bloom = BloomFilter(size, prob)

    for filepath in filepaths:
        print('loading {}...'.format(filepath))
        if pbarp:
            pbar = tqdm.tqdm(utils.openfile(filepath), ncols=100)
        else:
            pbar = utils.openfile(filepath)

        for item in csv.reader(pbar, delimiter=XSV_DELIMITER):
            try:
                token, count = item
                if token:
                    bloom.add(token)
                    if pbarp:
                        pbar.set_description(token)
            except:
                print('item: ', item)


    return bloom

if __name__ == '__main__':
    bloom = build_bloom(DEFAULT_DICTIONARY_FILES)
    word = input('> ')
    while word:
        print('இருக்குதா? {}'.format('இருக்கு' if word in bloom else 'இல்லை'))
        word = input('> ')
